''Roy Eltham's DHCP routines modified for use with the E1.31 DMX Bridge
''some comments added by jstj
''
'' DHCP  Version 1.0
'' Copyright (c) 2010 Roy Eltham
'' November 15, 2010
'' See end of file for terms of use

CON

' UDP port numbers for DHCP 
  DHCP_SERVER_PORT  = 67        ' from server to client
  DHCP_CLIENT_PORT  = 68        ' from client to server

  ' DHCP message OP code 
  DHCP_BOOTREQUEST  = 1
  DHCP_BOOTREPLY    = 2

  ' DHCP message type
  DHCP_DISCOVER     = 1
  DHCP_OFFER        = 2
  DHCP_REQUEST      = 3
  DHCP_DECLINE      = 4
  DHCP_ACK          = 5
  DHCP_NAK          = 6
  DHCP_RELEASE      = 7
  DHCP_INFORM       = 8

  ' MISC DHCP stuff
  DHCP_TIMEOUT        = 80_000_000 * 10  ' 10 seconds, assumes 80Mhz clkfreq
  DHCP_FLAGSBROADCAST = $8000
  DHCP_HTYPE10MB      = 1
  DHCP_HLENETHERNET   = 6
  MAGIC_COOKIE_0      = $63
  MAGIC_COOKIE_1      = $82
  MAGIC_COOKIE_2      = $53
  MAGIC_COOKIE_3      = $63
  
  ' DHCP Options
  OPT_padOption               = 0
  OPT_subnetMask              = 1
  OPT_timerOffset             = 2
  OPT_routersOnSubnet         = 3
  OPT_timeServer              = 4
  OPT_nameServer              = 5
  OPT_dns                     = 6
  OPT_logServer               = 7
  OPT_cookieServer            = 8
  OPT_lprServer               = 9
  OPT_impressServer           = 10
  OPT_resourceLocationServer  = 11
  OPT_hostName                = 12
  OPT_bootFileSize            = 13
  OPT_meritDumpFile           = 14
  OPT_domainName              = 15
  OPT_swapServer              = 16
  OPT_rootPath                = 17
  OPT_extentionsPath          = 18
  OPT_IPforwarding            = 19
  OPT_nonLocalSourceRouting   = 20
  OPT_policyFilter            = 21
  OPT_maxDgramReasmSize       = 22
  OPT_defaultIPTTL            = 23
  OPT_pathMTUagingTimeout     = 24
  OPT_pathMTUplateauTable     = 25
  OPT_ifMTU                   = 26
  OPT_allSubnetsLocal         = 27
  OPT_broadcastAddr           = 28
  OPT_performMaskDiscovery    = 29
  OPT_maskSupplier            = 30
  OPT_performRouterDiscovery  = 31
  OPT_routerSolicitationAddr  = 32
  OPT_staticRoute             = 33
  OPT_trailerEncapsulation    = 34
  OPT_arpCacheTimeout         = 35
  OPT_ethernetEncapsulation   = 36
  OPT_tcpDefaultTTL           = 37
  OPT_tcpKeepaliveInterval    = 38
  OPT_tcpKeepaliveGarbage     = 39
  OPT_nisDomainName           = 40
  OPT_nisServers              = 41
  OPT_ntpServers              = 42
  OPT_vendorSpecificInfo      = 43
  OPT_netBIOSnameServer       = 44
  OPT_netBIOSdgramDistServer  = 45
  OPT_netBIOSnodeType         = 46
  OPT_netBIOSscope            = 47
  OPT_xFontServer             = 48
  OPT_xDisplayManager         = 49
  OPT_dhcpRequestedIPaddr     = 50
  OPT_dhcpIPaddrLeaseTime     = 51
  OPT_dhcpOptionOverload      = 52
  OPT_dhcpMessageType         = 53
  OPT_dhcpServerIdentifier    = 54
  OPT_dhcpParamRequest        = 55
  OPT_dhcpMsg                 = 56
  OPT_dhcpMaxMsgSize          = 57
  OPT_dhcpT1value             = 58
  OPT_dhcpT2value             = 59
  OPT_dhcpClassIdentifier     = 60
  OPT_dhcpClientIdentifier    = 61
  OPT_endOption               = 255

  ' offsets for UDP header
  UDP_HEADER_DESTADDR     = 0
  UDP_HEADER_PORT         = 4
  UDP_HEADER_PAYLOADSIZE  = 6
  UDP_HEADER_SIZE         = 8

  ' offsets for DHCP packet (UDP header is in front of this)
  DHCP_OP     = 8
  DHCP_HTYPE  = 9
  DHCP_HLEN   = 10
  DHCP_HOPS   = 11
  DHCP_XID    = 12
  DHCP_SECS   = 16
  DHCP_FLAGS  = 18
  DHCP_CIADDR = 20
  DHCP_YIADDR = 24
  DHCP_SIADDR = 28
  DHCP_GIADDR = 32
  DHCP_CHADDR = 36
  DHCP_SNAME  = 52
  DHCP_FILE   = 116
  DHCP_COOKIE = 244
  DHCP_OPT    = 248
  DHCP_END    = 556
  
  ' return codes
  SUCCESS               = 0
  SOCKET_FAILED_OPENING = 1
  SERVER_TIMEOUT        = 2
  SERVER_NAK            = 3
  BUFFER_SIZE = 2048

  {  'WIZ module I/O
_reset          = 27              'WIZ module I/O here, this uses the SPI driver
_sclk           = 19              'WIZ module I/O serial clock to wiz module
_scs            = 22              'WIZ module I/O serial chip select
_miso           = 23              'WIZ module I/O master in slave out
_mosi           = 26              'WIZ module I/O master out slave in
}
VAR

  byte MAC_Address[6]
  byte IP[4]                    'IP, initially 169.254.74.73, later as assigned by DHCP, or set static by Web page             
  byte SubnetMask[4]            'set by DHCP or forced if static
  byte GatewayIP[4]             'set by DHCP or forced if static

  long DHCP_LeaseTime           'duration of DHCP lease in seconds, no renewal
  byte DHCP_Server[4]           'ip address of DHCP server
  byte DNS_Server[4]            'ip address of DNS server, set by DHCP, forced if static

  byte HostName[64]
  byte DomainName[64]           'you may need to make this larger if you have a very long domain name

  byte Buffer[BUFFER_SIZE]
  long cog
                                    
Obj

  WIZ:          "W5100_SPI_Driver.spin"    'Wiznet SPI mode driver from code repository                   

PUB Stop

  'If already running, stop cog
  if cog
    cogstop(cog~ -  1)

'' only call these GetXXX functions after calling DoDHCP()
PUB DoDHCP(dhcpsocket, XID) | dhcppacketSize,startCnt,timeoutCnt,currentCnt,serverReply,DestinationIP[4],status 
'    temporary variables used by this routine:
'    startcnt, timeoutcnt, and currentcnt are for the timeout timer
'    dhcppacketsize is the size of the received packet
'    serverreply is a status code returned by our parsedhcpreply routine: 0,1,2
'    destinationIP[4] is an array used to pass the destination IP value to the wiz socket open command
'    status  = socket status code read by a wiz call to readstatus

' first open the socket with the broadcast IP of 255.255.255.255, and the DHCP client port#, for UDP
  DestinationIP[0] := 255
  DestinationIP[1] := 255
  DestinationIP[2] := 255
  DestinationIP[3] := 255
  Wiz.SocketOpen(dhcpsocket, WIZ#_UDPPROTO, DHCP_CLIENT_PORT, DHCP_CLIENT_PORT, @DestinationIP[0])

  ' verify open was successful
  status := ReadStatus(dhcpsocket)
  if status <> Wiz#_SOCK_UDP
    return SOCKET_FAILED_OPENING    'means socket didn't open
   
  ' construct=send Discover packet
  FillHeader(255,255,255,255, 67, DHCP_END-UDP_HEADER_SIZE, @Buffer[0]) 'this builds the header portion
  FillDiscoverPacket(XID, @Buffer[0]) 'this builds everything else
  pausemsec(1000)    'this delay maybe could be shorter
  Wiz.txUDP(dhcpsocket, @Buffer[0])  'send it
  PauseMSec(100) 'pause a bit before looking for responses
     ' wait for a DHCP Offer reply
  startCnt := CNT                                       'this is the time when we started timing
  timeoutCnt := startCnt + DHCP_TIMEOUT                 'this will be the time that we timeout
  repeat                                                'do until we either timeout or get a packet
    dhcppacketsize := Wiz.rxUDP(dhcpsocket, @Buffer[0])    'get received packet size, if any
    if dhcppacketsize > 0
      serverReply := ParseDHCPReply(XID, DHCP_OFFER, @Buffer[0]) 'if we have one, then parse it and make reply code
      case serverReply                                  'then based on that reply code
        1:                                              '1 means we got our packet all is good, so quit this repeat
          QUIT
        2:                                              ' got NAK from server, so we bail with nak status
          Wiz.SocketClose(dhcpsocket)
          return SERVER_NAK
                                                        ' otherwise see if we have timed out yet, remember that
                                                        ' timeoutcnt can be less than startcnt if we wrap around
    currentCnt := CNT                                   'get time now
    if startCnt < timeoutCnt                            'this is true only if we are not wrapping around
      if ((currentCnt => timeoutCnt) OR (currentCnt =< startCnt)) ' made => and =<add parens ####if either true we have timed out
         return SERVER_TIMEOUT
    else                                                'this is if we are wrapping around
      if (currentCnt => timeoutCnt) AND (currentCnt =< startCnt) 'if both are true we've timed out
         return SERVER_TIMEOUT

  'we get here after we have received a reply message to our dhcp query
  'Now we will construct a DHCP Request packet and send it back to the server that
  'just replied to us with an offer, requesting to claim the offered IP
  FillHeader(255,255,255,255, 67, DHCP_END-UDP_HEADER_SIZE, @Buffer[0]) 'build the header portion of the req packet
  FillRequestPacket(XID, @Buffer[0])                 'build the rest of the request packet 
  Wiz.txUDP(dhcpsocket, @Buffer[0])                    'send the request packet
  'PauseMSec(100)                                        'short delay   #######
  ' now wait for a DHCP Ack reply, this grants us our claim on the IP
  startCnt := CNT
  timeoutCnt := startCnt + DHCP_TIMEOUT
  repeat
    dhcppacketsize := Wiz.rxUDP(dhcpsocket, @Buffer[0])    'check for a received packet
    if dhcppacketsize > 0                                    
       serverReply := ParseDHCPReply(XID, DHCP_ACK, @Buffer[0]) 'if we received one, parse it and build a status
                                                        'status 0= not for us, 1=got a good packet, 2 = got a NAK
      case serverReply
        1: ' we got our packet
           QUIT
        2: ' got NAK from server
          Wiz.SocketClose(dhcpsocket)
          return SERVER_NAK
          
    ' get here if we didn't get a good packet, now check for timeout, handling wrap case        
    currentCnt := CNT          
    if startCnt < timeoutCnt                            'if we won't be wrapping around
      if (currentCnt => timeoutCnt) OR (currentCnt =< startCnt) 'if either true we've timed out
         return SERVER_TIMEOUT
    else
      if (currentCnt => timeoutCnt) AND (currentCnt =< startCnt) 'if both true we've timed out
         return SERVER_TIMEOUT

    'get here after we've received and parsed a good ack packet, now write the relevant network values to the wiz:
  Wiz.SocketClose(0) 'close the socket  
  return SUCCESS            'done

PUB RenewLease(dhcpsocket)
  'still to be done
  return SERVER_NAK
  
PUB ReadStatus(dhcpsocket) | socketStatus
  Wiz.readSPI((Wiz#_S0_SR + (dhcpsocket * $0100)), @socketStatus, 1)
  return socketStatus

   'fills the header ortion of the outgoing packet buffer with dest IP address, port#, and payload size    
PRI FillHeader(ip0,ip1,ip2,ip3, port, payloadSize, pktbuf)
  byte[pktbuf][UDP_HEADER_DESTADDR] := ip3
  byte[pktbuf][UDP_HEADER_DESTADDR+1] := ip2
  byte[pktbuf][UDP_HEADER_DESTADDR+2] := ip1
  byte[pktbuf][UDP_HEADER_DESTADDR+3] := ip0
  byte[pktbuf][UDP_HEADER_PORT] := port.byte[1]
  byte[pktbuf][UDP_HEADER_PORT+1] := port.byte[0]
  byte[pktbuf][UDP_HEADER_PAYLOADSIZE] := payloadSize.byte[1]
  byte[pktbuf][UDP_HEADER_PAYLOADSIZE+1] := payloadSize.byte[0]

  'fills 1st part of packet after header, up to the option field, and the first few option sub-fields
  'returns with current offset value into the variable-length option field to be used by routine
  'that finished building the option field
PRI FillBasePacket(XID, messageType, pktbuf) | index, optionOffset, hostNameLength
  BYTEFILL(@byte[pktbuf][DHCP_OP], 0, DHCP_END-UDP_HEADER_SIZE)
  byte[pktbuf][DHCP_OP] := DHCP_BOOTREQUEST
  byte[pktbuf][DHCP_HTYPE] := DHCP_HTYPE10MB
  byte[pktbuf][DHCP_HLEN] := DHCP_HLENETHERNET
  byte[pktbuf][DHCP_XID+0] := XID.byte[3]
  byte[pktbuf][DHCP_XID+1] := XID.byte[2]
  byte[pktbuf][DHCP_XID+2] := XID.byte[1]
  byte[pktbuf][DHCP_XID+3] := XID.byte[0]
  'bug fix, this would cause an error in the creation of the DHCP packet if the packet buffer wasn't aligned
  'on a word boundary.  Previous line was:
  ' word[@byte[pktbuf][DHCP_FLAGS]] := ByteSwapWord(DHCP_FLAGSBROADCAST)
  'replaced with:
  byte[pktbuf][DHCP_FLAGS] := DHCP_FLAGSBROADCAST>>8  'high-order byte8
  byte[pktbuf][DHCP_FLAGS+1] := DHCP_FLAGSBROADCAST   'low-order byte
  'end bug fix
  repeat index from 0 to 5
    byte[pktbuf][DHCP_CHADDR+index] := MAC_Address[index]
    
  byte[pktbuf][DHCP_COOKIE+0] := MAGIC_COOKIE_0
  byte[pktbuf][DHCP_COOKIE+1] := MAGIC_COOKIE_1
  byte[pktbuf][DHCP_COOKIE+2] := MAGIC_COOKIE_2
  byte[pktbuf][DHCP_COOKIE+3] := MAGIC_COOKIE_3
  
  optionOffset := 0
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpMessageType
  byte[pktbuf][DHCP_OPT+optionOffset++] := 1
  byte[pktbuf][DHCP_OPT+optionOffset++] := messageType
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpClientIdentifier
  byte[pktbuf][DHCP_OPT+optionOffset++] := 7
  byte[pktbuf][DHCP_OPT+optionOffset++] := 1
  repeat index from 0 to 5
    byte[pktbuf][DHCP_OPT+optionOffset++] := MAC_Address[index]
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_hostName
  hostNameLength := STRSIZE(@HostName[0])
  ' force to be even, so next option after this will be word aligned (16bit)
  if (hostNameLength & 1) == 1
    hostNameLength += 1
  byte[pktbuf][DHCP_OPT+optionOffset++] := hostNameLength
  repeat index from 0 to hostNameLength-1
    byte[pktbuf][DHCP_OPT+optionOffset++] := HostName[index]
  
  return optionOffset
  
PRI FillDiscoverPacket(XID, pktbuf) | optionOffset
      'start by doing fillbasepacket, that returns with updated offset pointer into option field
  optionOffset := FillBasePacket(XID, DHCP_DISCOVER, pktbuf)
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpParamRequest
  byte[pktbuf][DHCP_OPT+optionOffset++] := 6
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_subnetMask
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_domainName
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dns
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_routersOnSubnet
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpT1value
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpT2value
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_endOption
  
PRI FillRequestPacket(XID, pktbuf) | optionOffset
  optionOffset := FillBasePacket(XID, DHCP_REQUEST, pktbuf)
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpRequestedIPaddr
  byte[pktbuf][DHCP_OPT+optionOffset++] := 4
  byte[pktbuf][DHCP_OPT+optionOffset++] := IP[0]
  byte[pktbuf][DHCP_OPT+optionOffset++] := IP[1]
  byte[pktbuf][DHCP_OPT+optionOffset++] := IP[2]
  byte[pktbuf][DHCP_OPT+optionOffset++] := IP[3]
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpServerIdentifier
  byte[pktbuf][DHCP_OPT+optionOffset++] := 4
  byte[pktbuf][DHCP_OPT+optionOffset++] := DHCP_Server[0]
  byte[pktbuf][DHCP_OPT+optionOffset++] := DHCP_Server[1]
  byte[pktbuf][DHCP_OPT+optionOffset++] := DHCP_Server[2]
  byte[pktbuf][DHCP_OPT+optionOffset++] := DHCP_Server[3]
  
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpParamRequest
  byte[pktbuf][DHCP_OPT+optionOffset++] := 8
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_subnetMask
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_routersOnSubnet
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dns
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_domainName
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpT1value
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_dhcpT2value
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_performRouterDiscovery
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_staticRoute
  byte[pktbuf][DHCP_OPT+optionOffset++] := OPT_endOption

PRI ParseDHCPReply(XID, expectedType, pktbuf) | index, optionOffset, option, optionLength, port, payloadSize
  
'bug fix, the previous code wouldn't work unless the packet buffer was aligned on a word boundary
' previous code was:
' port := ByteSwapWord(word[@byte[pktbuf][UDP_HEADER_PORT]])
' payloadSize := ByteSwapWord(word[@byte[pktbuf][UDP_HEADER_PAYLOADSIZE]])
' replaced with:
  port:=  byte[pktbuf][UDP_HEADER_PORT]*256 + byte[pktbuf][UDP_HEADER_PORT+1]
  payloadsize:= byte[pktbuf][UDP_HEADER_PAYLOADSIZE]*256 +  byte[pktbuf][UDP_HEADER_PAYLOADSIZE+1]
'end bug fix


   ' verify the packet is from the DHCP server and is at least of minimum size
  if port <> DHCP_SERVER_PORT OR payloadSize < DHCP_OPT
     return 0 ' this is not a DHCP packet

  ' verify that the packet is a DHCP Reply
  if byte[pktbuf][DHCP_OP] <> DHCP_BOOTREPLY
     return 0 ' this is not a DHCP Reply 
    
  ' verify XID match
  repeat index from 0 to 3
    if XID.byte[index] <> byte[pktbuf][DHCP_XID+3-index]
       return 0 ' does not match, so not for us
      
  ' verify MAC Address match
  repeat index from 0 to 5
    if MAC_Address[index] <> byte[pktbuf][DHCP_CHADDR+index]
      return 0 ' does not match, so not for us

  ' save the IP that is offered/claimed
  IP[0] := byte[pktbuf][DHCP_YIADDR+0]
  IP[1] := byte[pktbuf][DHCP_YIADDR+1]
  IP[2] := byte[pktbuf][DHCP_YIADDR+2]
  IP[3] := byte[pktbuf][DHCP_YIADDR+3]
  
  DHCP_LeaseTime := 0
  
  ' scan through options pulling out stuff we need
  optionOffset := 0
  repeat
    option := byte[pktbuf][DHCP_OPT+optionOffset++]
    if option <> 0
      optionLength := byte[pktbuf][DHCP_OPT+optionOffset++]
    else
      optionLength := 0
    case option
      OPT_endOption : 
        QUIT
      OPT_dhcpMessageType :
        if byte[pktbuf][DHCP_OPT+optionOffset] <> expectedType
          if byte[pktbuf][DHCP_OPT+optionOffset] == DHCP_NAK
            return 2
          else
            return 0
      OPT_subnetMask : 
        SubnetMask[0] := byte[pktbuf][DHCP_OPT+optionOffset+0]
        SubnetMask[1] := byte[pktbuf][DHCP_OPT+optionOffset+1]
        SubnetMask[2] := byte[pktbuf][DHCP_OPT+optionOffset+2]
        SubnetMask[3] := byte[pktbuf][DHCP_OPT+optionOffset+3]
      OPT_routersOnSubnet :
        GatewayIP[0] := byte[pktbuf][DHCP_OPT+optionOffset+0]
        GatewayIP[1] := byte[pktbuf][DHCP_OPT+optionOffset+1]
        GatewayIP[2] := byte[pktbuf][DHCP_OPT+optionOffset+2]
        GatewayIP[3] := byte[pktbuf][DHCP_OPT+optionOffset+3]
      OPT_dhcpIPaddrLeaseTime :
        DHCP_LeaseTime.byte[0] := byte[pktbuf][DHCP_OPT+optionOffset+3]
        DHCP_LeaseTime.byte[1] := byte[pktbuf][DHCP_OPT+optionOffset+2]
        DHCP_LeaseTime.byte[2] := byte[pktbuf][DHCP_OPT+optionOffset+1]
        DHCP_LeaseTime.byte[3] := byte[pktbuf][DHCP_OPT+optionOffset+0]
      OPT_dns :
        DNS_Server[0] := byte[pktbuf][DHCP_OPT+optionOffset+0]
        DNS_Server[1] := byte[pktbuf][DHCP_OPT+optionOffset+1]
        DNS_Server[2] := byte[pktbuf][DHCP_OPT+optionOffset+2]
        DNS_Server[3] := byte[pktbuf][DHCP_OPT+optionOffset+3]
      OPT_dhcpServerIdentifier :
        DHCP_Server[0] := byte[pktbuf][DHCP_OPT+optionOffset+0]
        DHCP_Server[1] := byte[pktbuf][DHCP_OPT+optionOffset+1]
        DHCP_Server[2] := byte[pktbuf][DHCP_OPT+optionOffset+2]
        DHCP_Server[3] := byte[pktbuf][DHCP_OPT+optionOffset+3]
      OPT_domainName :
        BYTEMOVE(@DomainName[0], @byte[pktbuf][DHCP_OPT+optionOffset], optionLength)
        
    optionOffset += optionLength
  
  return 1

PUB DHCPStart (HOST_NAME)
  ' clear all our variables at dhcp startup
BYTEFILL(@IP[0], 0, 152)   'note we don't use the mac_address variable from the dhcp code
  
  ' set the default host name to previously defined string, may be shorter
BYTEMOVE(@HostName[0], @HOST_NAME[0], STRSIZE(@HOST_NAME[0]))

   'because wiz has rev byte ordering from prop
PRI ByteSwapWord(value)
  result.byte[0] := value.byte[1]
  result.byte[1] := value.byte[0]
  
PRI PauseMSec(Duration)    'pause x milliseconds, use mult places
''  Pause execution for specified milliseconds.
''  This routine is based on the set clock frequency.
''  
''  params:  Duration = number of milliseconds to delay                                                                                               
''  return:  none
  waitcnt(((clkfreq / 1_000 * Duration - 3932) #> 381) + cnt)

DAT
{{
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                  
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
}}